# ruff: noqa: E401, E402, F401, F403
"""
!!! warning
    This module is deprecated and will be removed. Use
    `trulens.core.schema.feedback` or `trulens.core.schema.select` instead.
"""

from trulens.core.utils import deprecation as deprecation_utils

deprecation_utils.packages_dep_warn()

from trulens.core.schema.feedback import FeedbackCall
from trulens.core.schema.feedback import FeedbackCombinations
from trulens.core.schema.feedback import FeedbackDefinition
from trulens.core.schema.feedback import FeedbackMode
from trulens.core.schema.feedback import FeedbackOnMissingParameters
from trulens.core.schema.feedback import FeedbackResult
from trulens.core.schema.feedback import FeedbackResultStatus
from trulens.core.schema.select import Select



# ruff: noqa: E402, F403, F401
"""
!!! warning
    This module is deprecated and will be removed. Use
    `trulens.core.database.sqlalchemy` instead.
"""

from trulens.core.utils import deprecation as deprecation_utils

deprecation_utils.packages_dep_warn()

from trulens.core.database.sqlalchemy import AppsExtractor
from trulens.core.database.sqlalchemy import SnowflakeImpl
from trulens.core.database.sqlalchemy import SQLAlchemyDB
from trulens.core.database.sqlalchemy import _make_no_perf
from trulens.core.database.sqlalchemy import flatten

no_perf = _make_no_perf()



# ruff: noqa: E402
"""
!!! warning
    This module is deprecated and will be removed. Use `trulens.core.feedback`
    or `trulens.feedback` instead.
"""

from trulens.core.utils import deprecation as deprecation_utils

from trulens_eval._utils import optional as optional_utils

deprecation_utils.packages_dep_warn()

from trulens.core.utils import imports as import_utils

with import_utils.OptionalImports(messages=optional_utils.REQUIREMENT_FEEDBACK):
    from trulens.core.feedback.feedback import Feedback
    from trulens.feedback.embeddings import Embeddings
    from trulens.feedback.groundtruth import GroundTruthAgreement

with import_utils.OptionalImports(
    messages=optional_utils.REQUIREMENT_PROVIDER_LITELLM
):
    from trulens.providers.litellm.provider import LiteLLM

with import_utils.OptionalImports(
    messages=optional_utils.REQUIREMENT_PROVIDER_BEDROCK
):
    from trulens.providers.bedrock.provider import Bedrock

with import_utils.OptionalImports(
    messages=optional_utils.REQUIREMENT_PROVIDER_OPENAI
):
    from trulens.providers.openai.provider import AzureOpenAI
    from trulens.providers.openai.provider import OpenAI

with import_utils.OptionalImports(
    messages=optional_utils.REQUIREMENT_PROVIDER_HUGGINGFACE
):
    from trulens.providers.huggingface.provider import Huggingface
    from trulens.providers.huggingface.provider import HuggingfaceLocal

with import_utils.OptionalImports(
    messages=optional_utils.REQUIREMENT_PROVIDER_LANGCHAIN
):
    from trulens.providers.langchain.provider import Langchain

with import_utils.OptionalImports(
    messages=optional_utils.REQUIREMENT_PROVIDER_CORTEX
):
    from trulens.providers.cortex.provider import Cortex

__all__ = [
    "Feedback",
    "Embeddings",
    "GroundTruthAgreement",
    "OpenAI",
    "AzureOpenAI",
    "Huggingface",
    "HuggingfaceLocal",
    "LiteLLM",
    "Bedrock",
    "Langchain",
    "Cortex",
]

# Replace all classes we expose to ones which issue a deprecation warning upon
# initialization.
deprecation_utils.moved(
    globals(),
    names=__all__,
    old="trulens_eval.feedback",
    new="trulens.feedback",
)



# ruff: noqa: E402, F403, F401
"""
!!! warning
    This module is deprecated and will be removed. Use
    `trulens.core.database.migrations.data` instead.
"""

from trulens.core.utils import deprecation as deprecation_utils

deprecation_utils.packages_dep_warn()

from trulens.core.database.migrations.data import data_migrate
from trulens.core.database.migrations.data import sql_alchemy_migration_versions
from trulens.core.database.migrations.data import sqlalchemy_upgrade_paths



# ruff: noqa: E402, F401, F403
"""
!!! warning
    This module is deprecated and will be removed. Use `trulens.core.utils.trulens`
    instead.
"""

from trulens.core.utils import deprecation as deprecation_utils

deprecation_utils.packages_dep_warn()

from trulens.core.utils.trulens import *



# ruff: noqa: E401, E402, F401, F403
"""
!!! warning
    This module is deprecated and will be removed. Use
     `trulens.apps.nemo.tru_rails` instead.
"""

from trulens.core.utils import deprecation as deprecation_utils

deprecation_utils.packages_dep_warn()

from trulens.apps.nemo.tru_rails import FeedbackActions
from trulens.apps.nemo.tru_rails import RailsActionSelect
from trulens.apps.nemo.tru_rails import RailsInstrument
from trulens.apps.nemo.tru_rails import TruRails
from trulens.apps.nemo.tru_rails import registered_feedback_functions



# ruff: noqa: E401, E402, F401, F403
"""
!!! warning
    This module is deprecated and will be removed. Use
    `trulens.core.feedback.endpoint` or `trulens.feedback.dummy.endpoint`
    instead.
"""

from trulens.core.utils import deprecation as deprecation_utils

deprecation_utils.packages_dep_warn()

from trulens.core.feedback.endpoint import DEFAULT_RPM
from trulens.core.feedback.endpoint import INSTRUMENT
from trulens.core.feedback.endpoint import Endpoint
from trulens.core.feedback.endpoint import EndpointCallback
from trulens.core.utils.threading import DEFAULT_NETWORK_TIMEOUT
from trulens.feedback.dummy.endpoint import DummyEndpoint



# ruff: noqa: E402, F403, F401
"""
!!! warning
    This module is deprecated and will be removed. Use
    `trulens.core.database.orm` instead.
"""

from trulens.core.utils import deprecation as deprecation_utils

deprecation_utils.packages_dep_warn()

from trulens.core.database.orm import ORM
from trulens.core.database.orm import BaseWithTablePrefix
from trulens.core.database.orm import make_base_for_prefix
from trulens.core.database.orm import make_orm_for_prefix
from trulens.core.database.orm import new_base
from trulens.core.database.orm import new_orm



# ruff: noqa: E402, F401
"""
!!! warning
    This module is deprecated and will be removed. Use
    `trulens.core.utils.python` instead.
"""

from __future__ import annotations

from trulens.core.utils import deprecation as deprecation_utils

deprecation_utils.packages_dep_warn()
import dataclasses
from typing import (
    Any,
    Callable,
    Dict,
    Generic,
    Hashable,
    Optional,
    Type,
    TypeVar,
)

from trulens.core.utils.python import STACK
from trulens.core.utils.python import OpaqueWrapper
from trulens.core.utils.python import T
from trulens.core.utils.python import callable_name
from trulens.core.utils.python import caller_frame
from trulens.core.utils.python import caller_frameinfo
from trulens.core.utils.python import class_name
from trulens.core.utils.python import code_line
from trulens.core.utils.python import for_all_methods
from trulens.core.utils.python import get_all_local_in_call_stack
from trulens.core.utils.python import get_first_local_in_call_stack
from trulens.core.utils.python import get_task_stack
from trulens.core.utils.python import id_str
from trulens.core.utils.python import is_really_coroutinefunction
from trulens.core.utils.python import locals_except
from trulens.core.utils.python import logger
from trulens.core.utils.python import merge_stacks
from trulens.core.utils.python import module_name
from trulens.core.utils.python import run_before
from trulens.core.utils.python import safe_hasattr
from trulens.core.utils.python import safe_issubclass
from trulens.core.utils.python import safe_signature
from trulens.core.utils.python import stack_with_tasks
from trulens.core.utils.python import task_factory_with_stack
from trulens.core.utils.python import tru_new_event_loop
from trulens.core.utils.python import wrap_awaitable
from trulens.core.utils.python import wrap_generator


@dataclasses.dataclass
class SingletonInfo(Generic[T]):
    """
    Information about a singleton instance.
    """

    val: T
    """The singleton instance."""

    cls: Type[T]
    """The class of the singleton instance."""

    frameinfo_codeline: Optional[str]
    """The frame where the singleton was created.

    This is used for showing "already created" warnings. This is intentionally
    not the frame itself but a rendering of it to avoid maintaining references
    to frames and all of the things a frame holds onto.
    """

    name: Optional[str] = None
    """The name of the singleton instance.

    This is used for the SingletonPerName mechanism to have a separate singleton
    for each unique name (and class).
    """

    def __init__(self, name: str, val: Any):
        self.val = val
        self.cls = val.__class__
        self.name = name
        self.frameinfo_codeline = code_line(
            caller_frameinfo(offset=2), show_source=True
        )

    def warning(self):
        """Issue warning that this singleton already exists."""

        logger.warning(
            (
                "Singleton instance of type %s already created at:\n%s\n"
                "You can delete the singleton by calling `<instance>.delete_singleton()` or \n"
                f"""  ```python
  from trulens.core.utils.python import SingletonPerName
  SingletonPerName.delete_singleton_by_name(name="{self.name}", cls={self.cls.__name__})
  ```
            """
            ),
            self.cls.__name__,
            self.frameinfo_codeline,
        )


class SingletonPerName:
    """
    Class for creating singleton instances except there being one instance max,
    there is one max per different `name` argument. If `name` is never given,
    reverts to normal singleton behavior.
    """

    # Hold singleton instances here.
    _instances: Dict[Hashable, SingletonInfo[SingletonPerName]] = {}

    # Need some way to look up the name of the singleton instance. Cannot attach
    # a new attribute to instance since some metaclasses don't allow this (like
    # pydantic). We instead create a map from instance address to name.
    _id_to_name_map: Dict[int, Optional[str]] = {}

    def warning(self):
        """Issue warning that this singleton already exists."""

        name = SingletonPerName._id_to_name_map[id(self)]
        k = self.__class__.__name__, name
        if k in SingletonPerName._instances:
            SingletonPerName._instances[k].warning()
        else:
            raise RuntimeError(
                f"Instance of singleton type/name {k} does not exist."
            )

    def __new__(
        cls: Type[SingletonPerName],
        *args,
        name: Optional[str] = None,
        **kwargs,
    ) -> SingletonPerName:
        """
        Create the singleton instance if it doesn't already exist and return it.
        """

        k = cls.__name__, name

        if k not in cls._instances:
            logger.debug(
                "*** Creating new %s singleton instance for name = %s ***",
                cls.__name__,
                name,
            )
            # If exception happens here, the instance should not be added to
            # _instances.
            instance = super().__new__(cls)

            SingletonPerName._id_to_name_map[id(instance)] = name
            info: SingletonInfo = SingletonInfo(name=name, val=instance)
            SingletonPerName._instances[k] = info
        else:
            info = SingletonPerName._instances[k]
        obj = info.val
        assert isinstance(obj, cls)
        return obj

    @staticmethod
    def delete_singleton_by_name(
        name: str, cls: Optional[Type[SingletonPerName]] = None
    ):
        """
        Delete the singleton instance with the given name.

        This can be used for testing to create another singleton.

        Args:
            name: The name of the singleton instance to delete.

            cls: The class of the singleton instance to delete. If not given, all
                instances with the given name are deleted.
        """
        for k, v in list(SingletonPerName._instances.items()):
            if k[1] == name:
                if cls is not None and v.cls != cls:
                    continue

                del SingletonPerName._instances[k]
                del SingletonPerName._id_to_name_map[id(v.val)]

    def delete_singleton(self):
        """
        Delete the singleton instance. Can be used for testing to create another
        singleton.
        """
        id_ = id(self)

        if id_ in SingletonPerName._id_to_name_map:
            name = SingletonPerName._id_to_name_map[id_]
            del SingletonPerName._id_to_name_map[id_]
            del SingletonPerName._instances[(self.__class__.__name__, name)]
        else:
            logger.warning("Instance %s not found in our records.", self)



# ruff: noqa: E402, F401
"""
!!! warning
    This module is deprecated and will be removed. Use
    `trulens.core.utils.imports` instead.
"""

import warnings

from trulens.core.utils import deprecation as deprecation_utils

deprecation_utils.packages_dep_warn()

from trulens.core._utils.optional import (
    REQUIREMENT_APPS_LLAMA as REQUIREMENT_LLAMA,
)
from trulens.core._utils.optional import (
    REQUIREMENT_APPS_NEMO as REQUIREMENT_RAILS,
)
from trulens.core.utils.imports import MESSAGE_DEBUG_OPTIONAL_PACKAGE_NOT_FOUND
from trulens.core.utils.imports import MESSAGE_ERROR_REQUIRED_PACKAGE_NOT_FOUND
from trulens.core.utils.imports import MESSAGE_FRAGMENT_VERSION_MISMATCH
from trulens.core.utils.imports import (
    MESSAGE_FRAGMENT_VERSION_MISMATCH_OPTIONAL,
)
from trulens.core.utils.imports import MESSAGE_FRAGMENT_VERSION_MISMATCH_PIP
from trulens.core.utils.imports import (
    MESSAGE_FRAGMENT_VERSION_MISMATCH_REQUIRED,
)
from trulens.core.utils.imports import Dummy
from trulens.core.utils.imports import ImportErrorMessages
from trulens.core.utils.imports import OptionalImports
from trulens.core.utils.imports import VersionConflict
from trulens.core.utils.imports import (
    _requirements_of_trulens_core_file as requirements_of_file,
)
from trulens.core.utils.imports import all_packages
from trulens.core.utils.imports import check_imports
from trulens.core.utils.imports import format_import_errors
from trulens.core.utils.imports import get_package_version
from trulens.core.utils.imports import optional_packages
from trulens.core.utils.imports import parse_version
from trulens.core.utils.imports import pin_spec
from trulens.core.utils.imports import required_packages
from trulens.core.utils.imports import static_resource

from trulens_eval._utils.optional import REQUIREMENT_BERT_SCORE
from trulens_eval._utils.optional import REQUIREMENT_EVALUATE
from trulens_eval._utils.optional import REQUIREMENT_GROUNDEDNESS
from trulens_eval._utils.optional import REQUIREMENT_NOTEBOOK
from trulens_eval._utils.optional import REQUIREMENT_PINECONE
from trulens_eval._utils.optional import (
    REQUIREMENT_PROVIDER_BEDROCK as REQUIREMENT_BEDROCK,
)
from trulens_eval._utils.optional import (
    REQUIREMENT_PROVIDER_CORTEX as REQUIREMENT_CORTEX,
)
from trulens_eval._utils.optional import (
    REQUIREMENT_PROVIDER_HUGGINGFACE as REQUIREMENT_HUGGINGFACE,
)
from trulens_eval._utils.optional import (
    REQUIREMENT_PROVIDER_LANGCHAIN as REQUIREMENT_LANGCHAIN,
)
from trulens_eval._utils.optional import (
    REQUIREMENT_PROVIDER_LITELLM as REQUIREMENT_LITELLM,
)
from trulens_eval._utils.optional import (
    REQUIREMENT_PROVIDER_OPENAI as REQUIREMENT_OPENAI,
)
from trulens_eval._utils.optional import REQUIREMENT_SKLEARN

trulens_name: str = "trulens"



# ruff: noqa: E401, E402, F401, F403
"""
!!! warning
    This module is deprecated and will be removed. Use
     `trulens.core.session` or `trulens.dashboard.run` instead.
"""

from trulens.core.utils import deprecation as deprecation_utils

deprecation_utils.packages_dep_warn()

from trulens.core.session import TruSession as Tru
from trulens.core.utils import text as text_utils
from trulens.dashboard.run import DASHBOARD_START_TIMEOUT



# ruff: noqa: E402, F401
"""
!!! warning
    This module is deprecated and will be removed. Use `trulens.core.utils.containers`
    instead.
"""

from trulens.core.utils import deprecation as deprecation_utils

deprecation_utils.packages_dep_warn()

from trulens.core.utils.containers import BlockingSet
from trulens.core.utils.containers import dict_merge_with
from trulens.core.utils.containers import dict_set_with
from trulens.core.utils.containers import dict_set_with_multikey
from trulens.core.utils.containers import first
from trulens.core.utils.containers import is_empty
from trulens.core.utils.containers import iterable_peek
from trulens.core.utils.containers import second
from trulens.core.utils.containers import third



# ruff: noqa: E402, F401
"""
!!! warning
    This module is deprecated and will be removed. Use `trulens.core.utils.serial`
    instead.
"""

from trulens.core.utils import deprecation as deprecation_utils

deprecation_utils.packages_dep_warn()

from trulens.core.utils import serial as serial_utilsPath
from trulens.core.utils import serial as serial_utils_BASES
from trulens.core.utils import serial as serial_utilsized
from trulens.core.utils.serial import Collect
from trulens.core.utils.serial import GetAttribute
from trulens.core.utils.serial import GetIndex
from trulens.core.utils.serial import GetIndices
from trulens.core.utils.serial import GetItem
from trulens.core.utils.serial import GetItemOrAttribute
from trulens.core.utils.serial import GetItems
from trulens.core.utils.serial import GetSlice
from trulens.core.utils.serial import Lens
from trulens.core.utils.serial import ParseException
from trulens.core.utils.serial import SerialBytes
from trulens.core.utils.serial import SerialModel
from trulens.core.utils.serial import Step
from trulens.core.utils.serial import StepItemOrAttribute
from trulens.core.utils.serial import all_objects
from trulens.core.utils.serial import all_queries
from trulens.core.utils.serial import leaf_queries
from trulens.core.utils.serial import leafs
from trulens.core.utils.serial import matching_objects
from trulens.core.utils.serial import matching_queries
from trulens.core.utils.serial import model_dump



# ruff: noqa: E402
"""
!!! warning
    This module is deprecated and will be removed. Use `trulens.core`,
    `trulens.feedback`, `trulens.dashboard` instead.
"""

from trulens.core._utils import optional as optional_utils
from trulens.core.utils import deprecation as deprecation_utils
from trulens.core.utils import imports as import_utils

from trulens_eval._utils import optional as eval_optional_utils

deprecation_utils.packages_dep_warn()

__version_info__ = (1, 0, 0, "a")
__version__ = ".".join(map(str, __version_info__))

from trulens.apps.app import TruApp
from trulens.apps.basic import TruBasicApp
from trulens.apps.custom import TruCustomApp
from trulens.apps.virtual import TruVirtual
from trulens.core.feedback.feedback import Feedback
from trulens.core.feedback.provider import Provider
from trulens.core.schema import feedback as feedback_schema
from trulens.core.schema import select as select_schema
from trulens.core.session import TruSession as Tru
from trulens.core.utils.threading import TP

with import_utils.OptionalImports(
    messages=eval_optional_utils.REQUIREMENT_PROVIDER_LITELLM
):
    from trulens.providers.litellm.provider import LiteLLM

with import_utils.OptionalImports(
    messages=eval_optional_utils.REQUIREMENT_PROVIDER_BEDROCK
):
    from trulens.providers.bedrock.provider import Bedrock

with import_utils.OptionalImports(
    messages=eval_optional_utils.REQUIREMENT_PROVIDER_OPENAI
):
    from trulens.providers.openai.provider import AzureOpenAI
    from trulens.providers.openai.provider import OpenAI

with import_utils.OptionalImports(
    messages=eval_optional_utils.REQUIREMENT_PROVIDER_CORTEX
):
    from trulens.providers.cortex.provider import Cortex

with import_utils.OptionalImports(
    messages=eval_optional_utils.REQUIREMENT_PROVIDER_HUGGINGFACE
):
    from trulens.providers.huggingface.provider import Huggingface
    from trulens.providers.huggingface.provider import HuggingfaceLocal

with import_utils.OptionalImports(
    messages=eval_optional_utils.REQUIREMENT_PROVIDER_LANGCHAIN
):
    from trulens.providers.langchain.provider import Langchain

with import_utils.OptionalImports(
    messages=optional_utils.REQUIREMENT_APPS_LANGCHAIN
):
    from trulens.apps.langchain.tru_chain import TruChain

with import_utils.OptionalImports(
    messages=optional_utils.REQUIREMENT_APPS_LLAMA
):
    from trulens.apps.llamaindex.tru_llama import TruLlama

with import_utils.OptionalImports(
    messages=optional_utils.REQUIREMENT_APPS_NEMO
):
    from trulens.apps.nemo.tru_rails import TruRails

FeedbackMode = feedback_schema.FeedbackMode
Select = select_schema.Select

__all__ = [
    # main interface
    "Tru",
    # app types
    "TruBasicApp",
    "TruApp",
    "TruCustomApp",
    "TruChain",
    "TruLlama",
    "TruVirtual",
    "TruRails",
    # app setup
    "FeedbackMode",
    # feedback setup
    "Feedback",
    "Select",
    # feedback providers
    "Provider",
    "AzureOpenAI",
    "OpenAI",
    "Langchain",
    "LiteLLM",
    "Bedrock",
    "Huggingface",
    "HuggingfaceLocal",
    "Cortex",
    # misc utility
    "TP",
]

# Replace all classes we expose to ones which issue a deprecation warning upon
# initialization.
deprecation_utils.moved(globals(), names=__all__)
